@@ -33,7 +33,7 @@ please cite the paper: http://arxiv.org/abs/1305.4194
 
 TODOs:
 
-  For 2.1.3:
+  For 2.1.4:
    * In set_value, if first item is non-int, treat as None-indexed
    * Refine value string heuristic for strings with ints in them?
 
@@ -43,13 +43,21 @@ TODOs:
      separation). Can use separate comment dicts in Block and Decay, and
      attach a multiline .comment attr to the returned/written dicts.
 
+  For 3.0.0:
+   * Add a new Doc type which encapsulates the blocks, decays, xsections,
+   etc. data sections and will act as a more robust single return value from
+   the read commands / argument to the write commands. Can also handle
+   document-level header comments this way and improve the __repr__ resulting
+   from an interactive read() call without a target variable (i.e. hide the
+   uninteresting OrderedDict boilerplate).
+
   Later/maybe:
    * Identify HERWIG decay matrix element to use in ISAWIG
    * Handle RPV SUSY in ISAWIG
 """
 
 __author__ = "Andy Buckley <andy.buckley@cern.ch>"
-__version__ = "2.1.2"
+__version__ = "2.1.3"
 
 
 
@@ -344,13 +352,16 @@ class Particle(object):
 ###############################################################################
 ## SLHA parsing and writing functions
 
-def readSLHA(spcstr, ignorenobr=False):
+def readSLHA(spcstr, ignorenobr=False, ignorenomass=False):
     """
     Read an SLHA definition from a string, returning dictionaries of blocks and
     decays.
 
     If the ignorenobr parameter is True, do not store decay entries with a
     branching ratio of zero.
+
+    If the ignorenomass parameter is True, parse file even if mass block is
+    absent in the file (default is to raise a ParseError).
     """
     blocks = _mkdict()
     decays = _mkdict()
@@ -365,37 +376,55 @@ def readSLHA(spcstr, ignorenobr=False):
             continue
         if "#" in line:
             line = line[:line.index("#")]
+        ## Ignore empty lines (after comment removal and whitespace trimming)
+        if not line.strip():
+            continue
 
-        ## Handle BLOCK/DECAY start lines
-        if line.upper().startswith("BLOCK"):
-            #print line
-            match = re.match(r"BLOCK\s+(\w+)(\s+Q\s*=\s*.+)?", line.upper())
-            if not match:
-                continue
-            blockname = match.group(1)
-            qstr = match.group(2)
-            if qstr is not None:
-                qstr = qstr[qstr.find("=")+1:].strip()
-            currentblock = blockname
-            currentdecay = None
-            blocks[blockname] = Block(blockname, q=qstr)
-        elif line.upper().startswith("DECAY"):
-            match = re.match(r"DECAY\s+(-?\d+)\s+([\d\.E+-]+|NAN).*", line.upper())
-            if not match:
-                continue
-            pdgid = int(match.group(1))
-            width = float(match.group(2)) if match.group(2) != "NAN" else None
-            currentblock = "DECAY"
-            currentdecay = pdgid
-            decays[pdgid] = Particle(pdgid, width)
+        ## Section header lines start with a non-whitespace character, data lines have a whitespace indent
+        # TODO: Are tabs also allowed for indents? Check the SLHA standard.
+        if not line.startswith(" "):
+            # TODO: Should we now strip the line to remove any trailing whitespace?
+
+            ## Handle BLOCK start lines
+            if line.upper().startswith("BLOCK"):
+                #print line
+                match = re.match(r"BLOCK\s+(\w+)(\s+Q\s*=\s*.+)?", line.upper())
+                if not match:
+                    continue
+                blockname = match.group(1)
+                qstr = match.group(2)
+                if qstr is not None:
+                    qstr = qstr[qstr.find("=")+1:].strip()
+                currentblock = blockname
+                currentdecay = None
+                blocks[blockname] = Block(blockname, q=qstr)
+            ## Handle DECAY start lines
+            elif line.upper().startswith("DECAY"):
+                match = re.match(r"DECAY\s+(-?\d+)\s+([\d\.E+-]+|NAN).*", line.upper())
+                if not match:
+                    continue
+                pdgid = int(match.group(1))
+                width = float(match.group(2)) if match.group(2) != "NAN" else None
+                currentblock = "DECAY"
+                currentdecay = pdgid
+                decays[pdgid] = Particle(pdgid, width)
+            ## Handle unknown section type start lines (and continue ignoring until a non-header line is found)
+            elif type(_autotype(line.split()[0])) is str:
+                import sys
+                sys.stderr.write("Ignoring unknown section type: %s\n" % line.split()[0])
+                currentblock = None
+                currentdecay = None
+
+        ## This non-empty line starts with an indent, hence must be an in-section data line
         else:
-            ## In-block line
+            # TODO: Should we now strip the line to remove the indent (and any trailing whitespace)?
             if currentblock is not None:
                 items = line.split()
                 if len(items) < 1:
                     continue
                 if currentblock != "DECAY":
                     blocks[currentblock].add_entry(items)
+                # TODO: Add handling of XSECTION if/when standardised
                 else:
                     br = float(items[0]) if items[0].upper() != "NAN" else None
                     nda = int(items[1])
@@ -410,7 +439,8 @@ def readSLHA(spcstr, ignorenobr=False):
             if decays.has_key(pid):
                 decays[pid].mass = blocks["MASS"][pid]
     except:
-        raise ParseError("No MASS block found: cannot populate particle masses")
+        if not ignorenomass:
+            raise ParseError("No MASS block found: cannot populate particle masses")
 
     return blocks, decays
 
@@ -1069,7 +1099,7 @@ def readSLHAFile(spcfilename, **kwargs):
     Other keyword parameters are passed to readSLHA.
     """
     f = open(spcfilename, "r")
-    rtn = readSLHA(f.read(), kwargs)
+    rtn = readSLHA(f.read(), **kwargs)
     f.close()
     return rtn
 
@@ -1081,7 +1111,7 @@ def writeSLHAFile(spcfilename, blocks, decays, **kwargs):
     Other keyword parameters are passed to writeSLHA.
     """
     f = open(spcfilename, "w")
-    f.write(writeSLHA(blocks, decays, kwargs))
+    f.write(writeSLHA(blocks, decays, **kwargs))
     f.close()
 
 
@@ -1095,7 +1125,7 @@ def readISAWIGFile(isafilename, **kwargs):
     Other keyword parameters are passed to readSLHA.
     """
     f = open(isafilename, "r")
-    rtn = readISAWIG(f.read(), kwargs)
+    rtn = readISAWIG(f.read(), **kwargs)
     f.close()
     return rtn
 
@@ -1107,7 +1137,7 @@ def writeISAWIGFile(isafilename, blocks, decays, **kwargs):
     Other keyword parameters are passed to writeISAWIG.
     """
     f = open(isafilename, "w")
-    f.write(writeISAWIG(blocks, decays, kwargs))
+    f.write(writeISAWIG(blocks, decays, **kwargs))
     f.close()
 
 
